package com.ftdi.uart2spi.ui;

import java.awt.EventQueue;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.DefaultCaret;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import javax.swing.JTextPane;
import javax.swing.JLabel;
import javax.swing.JComboBox;
import javax.swing.JSeparator;
import javax.swing.JButton;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

import java.awt.Font;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import javax.swing.JTextField;

import com.ftdi.FT90xSPI;
import com.ftdi.swing.IconStore;
import com.ftdi.swing.JNonWrappingTextPane;
import com.ftdi.util.AppProperties;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;

import javax.swing.SwingConstants;
import javax.swing.UIManager;

public class MainUI extends JFrame {

	private JPanel contentPane;
	
	private static final String PARSE_TYPES[] = {"Text","Decimal","Hexadecimal"};
	private static final String COPYRIGHT_STRING = AppProperties.getProperties().getProperty("ftdi.copyright");
	
	private final Icon iconWindow = IconStore.LoadIcon("/gfx/ftdiico.png");
	
	private final ArrayList<byte[]> receivedBytes = new ArrayList<byte[]>();
	private final ReceiveActionListener receivedListener;
	private final TransmitActionListener transmitListener;

	/**
	 * Launch the application.
	 */
	public static void main(String[] args) {
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				try {
					MainUI frame = new MainUI("Test", null);
					frame.setVisible(true);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		});
	}

	/**
	 * Create the frame.
	 */
	public MainUI(String title, FT90xSPI dev) {
		
		setResizable(false);
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		setBounds(100, 100, 434, 542);
		contentPane = new JPanel();
		contentPane.setBorder(new EmptyBorder(5, 5, 5, 5));
		setTitle(title);
		setContentPane(contentPane);
		contentPane.setLayout(null);
		setLocationRelativeTo(null);

		setIconImage(((ImageIcon)iconWindow).getImage());
		
		JLabel lblTransmit = new JLabel("Transmit:");
		lblTransmit.setFont(lblTransmit.getFont().deriveFont(lblTransmit.getFont().getStyle() | Font.BOLD, lblTransmit.getFont().getSize() + 4f));
		lblTransmit.setBounds(10, 8, 97, 23);
		contentPane.add(lblTransmit);
		
		JComboBox<String> comboBoxTXParse = new JComboBox<String>();
		comboBoxTXParse.setBounds(346, 11, 72, 20);
		for (String s:PARSE_TYPES)
			comboBoxTXParse.addItem(s);
		contentPane.add(comboBoxTXParse);
		
		JLabel lblReceive = new JLabel("Receive:");
		lblReceive.setFont(lblReceive.getFont().deriveFont(lblReceive.getFont().getStyle() | Font.BOLD, lblReceive.getFont().getSize() + 4f));
		lblReceive.setBounds(10, 275, 97, 17);
		contentPane.add(lblReceive);
		
		JComboBox<String> comboBoxRXParse = new JComboBox<String>();
		comboBoxRXParse.setBounds(346, 275, 72, 20);
		for (String s:PARSE_TYPES)
			comboBoxRXParse.addItem(s);
		contentPane.add(comboBoxRXParse);
		
		JButton btnTransmit = new JButton("Transmit");
		btnTransmit.setBounds(10, 225, 89, 23);
		contentPane.add(btnTransmit);
		
		JButton btnClear = new JButton("Clear");
		btnClear.setBounds(264, 225, 72, 23);
		contentPane.add(btnClear);
		
		JButton btnLoad = new JButton("Load...");
		btnLoad.setBounds(346, 225, 72, 23);
		contentPane.add(btnLoad);
		
		JScrollPane scrollPaneTX = new JScrollPane();
		scrollPaneTX.setBounds(10, 39, 408, 175);
		contentPane.add(scrollPaneTX);
		
		final JTextPane textPaneTX = new JNonWrappingTextPane();
		textPaneTX.setSelectionColor(Color.GREEN);
		textPaneTX.setSelectedTextColor(Color.BLACK);
		textPaneTX.setCaretColor(Color.GREEN);
		textPaneTX.setForeground(Color.GREEN);
		textPaneTX.setBackground(Color.BLACK);
		textPaneTX.setFont(textPaneTX.getFont().deriveFont(textPaneTX.getFont().getStyle() | Font.BOLD, textPaneTX.getFont().getSize() + 2f));
		scrollPaneTX.setViewportView(textPaneTX);
		
		JScrollPane scrollPaneRX = new JScrollPane();
		scrollPaneRX.setBounds(10, 303, 408, 175);
		contentPane.add(scrollPaneRX);
		
		JTextPane textPaneRX = new JNonWrappingTextPane();
		textPaneRX.setSelectionColor(Color.YELLOW);
		textPaneRX.setSelectedTextColor(Color.BLACK);
		textPaneRX.setForeground(Color.YELLOW);
		textPaneRX.setBackground(Color.BLACK);
		textPaneRX.setFont(textPaneRX.getFont().deriveFont(textPaneRX.getFont().getStyle() | Font.BOLD, textPaneRX.getFont().getSize() + 2f));
		textPaneRX.setEditable(false);
		scrollPaneRX.setViewportView(textPaneRX);
		
		JSeparator separator = new JSeparator();
		separator.setBounds(0, 259, 428, 2);
		contentPane.add(separator);
		
		final JTextField textFieldStatusBar = new JTextField();
		textFieldStatusBar.setText(COPYRIGHT_STRING);
		textFieldStatusBar.setEditable(false);
		textFieldStatusBar.setBounds(0, 494, 428, 20);
		contentPane.add(textFieldStatusBar);
		
		JLabel lblComPort = new JLabel("");
		if (dev != null)
		{
			lblComPort.setText("Using "+dev.getPort());
		}
		lblComPort.setBounds(109, 229, 145, 14);
		contentPane.add(lblComPort);
		
		JLabel lblFormat1 = new JLabel("Format:");
		lblFormat1.setHorizontalAlignment(SwingConstants.TRAILING);
		lblFormat1.setBounds(295, 14, 46, 14);
		contentPane.add(lblFormat1);
		
		JLabel lblFormat2 = new JLabel("Format:");
		lblFormat2.setHorizontalAlignment(SwingConstants.TRAILING);
		lblFormat2.setBounds(295, 278, 46, 14);
		contentPane.add(lblFormat2);
		
		/* Add transmit listener */
		transmitListener = new TransmitActionListener(textPaneTX, comboBoxTXParse, textFieldStatusBar, dev);
		btnTransmit.addActionListener(transmitListener);
		
		/* Add receive listener */
		receivedListener = new ReceiveActionListener(textPaneRX, comboBoxRXParse, textFieldStatusBar);
		comboBoxRXParse.addActionListener(receivedListener);

		/* Add clear action */
		btnClear.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) { textPaneTX.setText(null); }
		});

		/* Add load from file action */
		btnLoad.addActionListener(new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser fc = new JFileChooser();
				
				/* Add a filter for text file */
				FileNameExtensionFilter ffiltText = new FileNameExtensionFilter("Text Files (*.txt)", "txt");
				fc.addChoosableFileFilter(ffiltText);
				fc.setFileFilter(ffiltText);
				
				fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				
				fc.setCurrentDirectory(new File(AppProperties.getProperties().getProperty("user.dir")));

				int ret = fc.showOpenDialog(null);
				
				if (ret == JFileChooser.APPROVE_OPTION)
				{
					File f = fc.getSelectedFile();
					try {
						System.out.println("Opening "+f);
						FileReader fr = new FileReader(f);
						BufferedReader br = new BufferedReader(fr);
						textPaneTX.read(br, null);
						br.close();
					} catch (FileNotFoundException ex) {
						System.out.println("Cannot find file "+f);
						textFieldStatusBar.setForeground(Color.RED);
						textFieldStatusBar.setText("Error: Cannot find file "+f);
					} catch (IOException ex) {
						ex.printStackTrace();
					}
					
				}
			}
		});
		
	}
	
	public void showAndWaitForClose()
	{
		setVisible(true);
		while (isVisible());
	}
	
	
	
	/** Transmit Listener
	 */
	private class TransmitActionListener implements ActionListener
	{
		private final JTextPane tp;
		private final JComboBox<String> type;
		private final JTextField error;
		private final FT90xSPI spi;
		public TransmitActionListener(JTextPane tp, JComboBox<String> type, JTextField error, FT90xSPI spi) 
		{
			this.tp = tp;
			this.type = type;
			this.error = error;
			this.spi = spi;
		}
		@Override
		public void actionPerformed(ActionEvent e) 
		{
			new Thread(new Runnable() {
				@Override
				public void run() 
				{
					String parse = (String)type.getSelectedItem();
					String text = tp.getText();

					/* Shove in some default text */
					error.setForeground(Color.BLACK);
					error.setText(COPYRIGHT_STRING);
					
					try
					{
						receivedBytes.clear();
						
						if ((text != null) && (text.length() > 0))
						{
							for (String line : text.split("\r\n"))
							{
								System.out.println("Transferring \""+line+"\" as "+parse);
								
								byte[] data = parse(line, parse);
								
								if (data.length > 255)
								{
									throw new TooMuchDataException();
								}
								
								if (spi != null)
								{
									receivedBytes.add(spi.transfer(data));
								}
								else
								{
									/* Testing: Loopback */
									receivedBytes.add(data);
								}
									
							}
						}
					}
					catch (NumberFormatException ex)
					{
						receivedBytes.clear();
						error.setForeground(Color.RED);
						error.setText("Error: Unknown Number. "+ex.getMessage());
					}
					catch (UnknownArgumentException ex)
					{
						receivedBytes.clear();
						error.setForeground(Color.RED);
						error.setText("Error: Unknown Argument. "+parse);
					}
					catch (TooMuchDataException ex)
					{
						receivedBytes.clear();
						error.setForeground(Color.RED);
						error.setText("Error: Too many bytes in a transfer");
					}
					catch (Exception ex)
					{
						receivedBytes.clear();
						ex.printStackTrace();
						error.setText(ex.toString());
					}
					finally
					{
						receivedListener.actionPerformed(null);
					}
				}
			}).start();
		}
		
		private byte[] parse(String line, String parseTo)
			throws NumberFormatException, UnknownArgumentException
		{
			byte[] value;
					
			if (parseTo == PARSE_TYPES[0] /* Text */)
			{
				value = line.getBytes();
			}
			else if ((parseTo == PARSE_TYPES[1]) /* Decimal */ ||
					(parseTo == PARSE_TYPES[2]) /* Hex */)
			{
				int radix;
				
				if (parseTo == PARSE_TYPES[1]) 
					radix = 10;
				else if (parseTo == PARSE_TYPES[2]) 
					radix = 16;
				else throw new UnknownArgumentException();

				String[] words = line.split("\\s+");
				
				value = new byte[words.length];
				
				for (int i = 0; i < words.length; ++i)
				{
					int j = Integer.parseInt(words[i], radix);
										
					if (j > 255) throw new NumberFormatException("For input string \""+words[i]+"\"");
					if (j > Byte.MAX_VALUE) j -= 256;
					
					if (j > Byte.MAX_VALUE) throw new NumberFormatException("For input string \""+words[i]+"\"");
					if (j < Byte.MIN_VALUE) throw new NumberFormatException("For input string \""+words[i]+"\"");
					
					value[i] = (byte) j;	
				}
			}
			else
			{
				throw new UnknownArgumentException();
			}
			
			return value;
				
		}
		
		private class UnknownArgumentException extends Exception {}
		private class TooMuchDataException extends Exception {}
	}
	
	/** Receive Listener
	 */
	private class ReceiveActionListener implements ActionListener
	{
		private JTextPane tp;
		private JComboBox<String> type;
		private JTextField error;
		public ReceiveActionListener(JTextPane tp, JComboBox<String> type, JTextField error) 
		{
			this.tp = tp;
			this.type = type;
			this.error = error;
		}
		@Override
		public void actionPerformed(ActionEvent arg0) 
		{
			new Thread(new Runnable() {
				@Override
				public void run() 
				{
					StringBuilder sb = new StringBuilder();
					String parseTo = (String)type.getSelectedItem();
										
					try 
					{
						for (byte[] transfer : receivedBytes)
						{

							/*System.out.print("rx -> ");
							for (byte b : transfer) System.out.print(b+",");
							System.out.println();*/
							
							sb.append(parse(transfer, parseTo));
							sb.append("\r\n");
						}
						tp.setText(sb.toString());
					}
					catch (UnknownArgumentException ex)
					{
						tp.setText("");
					}
				}
			}).start();
		}
		
		private String parse(byte[] data, String parseTo)
			throws UnknownArgumentException
		{
			StringBuilder sb = new StringBuilder();
			
			if (parseTo == PARSE_TYPES[0] /* Text */)
			{
				for (byte b : data)
					sb.append((char)b);
			}
			else if (parseTo == PARSE_TYPES[1] /* Decimal */)
			{
				for (int i = 0; i < data.length; ++i)
				{
					//int j = data[i];
					//if (j < 0) j += 256;
					//sb.append(j);
					sb.append(data[i] < 0 ? data[i] + 256 : data[i]);
					if (i < (data.length - 1)) sb.append(" ");
				}
			}
			else if (parseTo == PARSE_TYPES[2] /* Hex */)
			{
				for (int i = 0; i < data.length; ++i)
				{
					int j = data[i];
					String s = String.format("%02X", j);
					//s = s.substring(s.length()-2); /* Only take the lowest two nibbles */
					//sb.append(s);
					sb.append(s.substring(s.length()-2));
					if (i < (data.length - 1)) sb.append(" ");
				}		
			}
			else
			{
				throw new UnknownArgumentException();
			}
			
			return sb.toString();
		}
		
		private class UnknownArgumentException extends Exception {}
	}
}
